cimport cython
from libc.stdint cimport int32_t

import numpy as np


def nearest_nonoverlapping(prev_ridx, prev_dist,
                           next_ridx, next_dist):

    if prev_dist.dtype == np.int32:
        return nearest_nonoverlapping32(prev_ridx, prev_dist,
                                        next_ridx, next_dist)
    elif prev_dist.dtype == np.int64:
        return nearest_nonoverlapping64(prev_ridx, prev_dist,
                                        next_ridx, next_dist)
    else:
        raise Exception("Starts/Ends not int64 or int32: " + str(prev_dist.dtype))

def nearest_previous_nonoverlapping(ls, re, ridx):

    if ls.dtype == np.int32:
        return nearest_previous_nonoverlapping32(ls, re, ridx)
    elif ls.dtype == np.int64:
        return nearest_previous_nonoverlapping64(ls, re, ridx)
    else:
        raise Exception("Starts/Ends not int64 or int32: " + str(ls.dtype))


def nearest_next_nonoverlapping(le, rs, ridx):

    if le.dtype == np.int32:
        return nearest_next_nonoverlapping32(le, rs, ridx)
    elif le.dtype == np.int64:
        return nearest_next_nonoverlapping64(le, rs, ridx)
    else:
        raise Exception("Starts/Ends not int64 or int32: " + str(le.dtype))


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_next_nonoverlapping64(const long [::1] l_e, const long [::1] r_s, const long [::1] r_idx):

    cdef int j = 0
    cdef int i = 0

    cdef int len_l = len(l_e)
    cdef int len_r = len(r_s)

    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int_) * -1
    cdef long [::1] ridx
    cdef long [::1] dist

    ridx = arr_ridx
    dist = arr_dist

    # print("l_e", list(l_e))
    # print("r_s", list(r_s))
    while i < len_l and j < len_r:
        # print("l_e[i]", l_e[i])
        # print("r_s[j]", r_s[j])
        if l_e[i] >= r_s[j]:
            # print("in if")
            j += 1
        else:
            # print("in else")
            dist[i] = r_s[j] - l_e[i]
            ridx[i] = r_idx[j]
            i += 1
        # print("dist", list(dist))
        # print("ridx", list(ridx))

    return arr_ridx, arr_dist

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_next_nonoverlapping32(const int32_t [::1] l_e, const int32_t [::1] r_s, const long [::1] r_idx):

    cdef int j = 0
    cdef int i = 0

    cdef int len_l = len(l_e)
    cdef int len_r = len(r_s)

    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int32) * -1
    cdef long [::1] ridx
    cdef int32_t [::1] dist

    ridx = arr_ridx
    dist = arr_dist

    # print("l_e", list(l_e))
    # print("r_s", list(r_s))
    while i < len_l and j < len_r:
        # print("l_e[i]", l_e[i])
        # print("r_s[j]", r_s[j])
        if l_e[i] >= r_s[j]:
            # print("in if")
            j += 1
        else:
            # print("in else")
            dist[i] = r_s[j] - l_e[i]
            ridx[i] = r_idx[j]
            i += 1
            # print("dist", list(dist))
            # print("ridx", list(ridx))

    return arr_ridx, arr_dist



@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_previous_nonoverlapping64(const long [::1] l_s, const long [::1] r_e, const long [::1] r_idx):

    cdef int len_l = len(l_s)
    cdef int len_r = len(r_e)

    cdef int j = len_r - 1
    cdef int i = len_l - 1

    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int_) * -1
    cdef long [::1] ridx
    cdef long [::1] dist

    ridx = arr_ridx
    dist = arr_dist

    while -1 < i and -1 < j:
        if l_s[i] <= r_e[j]:
            # print("in if")
            j -= 1
        else:
            # print("in else")
            # print("l_s[i]", l_s[i])
            # print("r_e[j]", r_e[j])
            dist[i] = l_s[i] - r_e[j]
            ridx[i] = r_idx[j]
            i -= 1
        # print("dist", list(dist))
        # print("ridx", list(ridx))

    # print("final dist", list(dist))
    # print("final ridx", list(ridx))

    return arr_ridx, arr_dist

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_previous_nonoverlapping32(const int32_t [::1] l_s, const int32_t [::1] r_e, const long [::1] r_idx):

    cdef int len_l = len(l_s)
    cdef int len_r = len(r_e)

    cdef int j = len_r - 1
    cdef int i = len_l - 1

    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int32) * -1
    cdef long [::1] ridx
    cdef int32_t [::1] dist

    ridx = arr_ridx
    dist = arr_dist

    while -1 < i and -1 < j:
        if l_s[i] <= r_e[j]:
            # print("in if")
            j -= 1
        else:
            # print("in else")
            # print("l_s[i]", l_s[i])
            # print("r_e[j]", r_e[j])
            dist[i] = l_s[i] - r_e[j]
            ridx[i] = r_idx[j]
            i -= 1
            # print("dist", list(dist))
            # print("ridx", list(ridx))

    # print("final dist", list(dist))
    # print("final ridx", list(ridx))

    return arr_ridx, arr_dist

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_nonoverlapping64(const long [::1] prev_ridx, const long [::1] prev_dist,
                               const long [::1] next_ridx, const long [::1] next_dist):

    cdef int i = 0

    cdef int length = len(prev_ridx)

    output_arr_ridx = np.ones(length, dtype=np.int_) * -1
    output_arr_dist = np.ones(length, dtype=np.int_) * -1

    cdef long [::1] output_ridx
    cdef long [::1] output_dist

    output_ridx = output_arr_ridx
    output_dist = output_arr_dist

    for i in range(length):
        if next_ridx[i] == -1 and prev_ridx[i] > -1:
            output_ridx[i] = prev_ridx[i]
            output_dist[i] = prev_dist[i]
        elif prev_ridx[i] > -1 and next_ridx[i] > -1 and prev_dist[i] <= next_dist[i]:
            output_ridx[i] = prev_ridx[i]
            output_dist[i] = prev_dist[i]
        elif next_dist[i] > -1:
            output_ridx[i] = next_ridx[i]
            output_dist[i] = next_dist[i]

    return output_arr_ridx, output_arr_dist


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_nonoverlapping32(const long [::1] prev_ridx, const int32_t [::1] prev_dist,
                               const long [::1] next_ridx, const int32_t [::1] next_dist):

    cdef int i = 0

    cdef int length = len(prev_ridx)

    output_arr_ridx = np.ones(length, dtype=np.int_) * -1
    output_arr_dist = np.ones(length, dtype=np.int32) * -1

    cdef long [::1] output_ridx
    cdef int32_t [::1] output_dist

    output_ridx = output_arr_ridx
    output_dist = output_arr_dist

    for i in range(length):
        if next_ridx[i] == -1 and prev_ridx[i] > -1:
            output_ridx[i] = prev_ridx[i]
            output_dist[i] = prev_dist[i]
        elif prev_ridx[i] > -1 and next_ridx[i] > -1 and prev_dist[i] <= next_dist[i]:
            output_ridx[i] = prev_ridx[i]
            output_dist[i] = prev_dist[i]
        elif next_dist[i] > -1:
            output_ridx[i] = next_ridx[i]
            output_dist[i] = next_dist[i]

    return output_arr_ridx, output_arr_dist



from io import StringIO

import numpy as np
import pandas as pd
import pytest

from sorted_nearest import get_all_ties, get_different_ties


@pytest.fixture
def data():
    c = """ids dist
1 1
1 1
1 2
1 3
0 5000
0 5000
0 5000
2 100
2 110
2 110
2 111
3 111
3 111
3 111
3 112
3 112
3 113
4 112
4 113
4 113
4 113
4 113
4 113
4 113
4 150"""

    df = pd.read_table(StringIO(c), header=0, sep="\s+")

    return df


def test_get_all_ties(data):
    df = data

    print(df)

    k = 2
    result = get_all_ties(df.index.values, df.ids.values, df.dist.values, k)
    print(df.reindex(result))
    print(result)

    expected = [0, 1, 4, 5, 6, 7, 8, 9, 11, 12, 13, 17, 18, 19, 20, 21, 22, 23]
    assert list(result) == expected


def test_get_different_ties(data):
    df = data

    k = 2
    result = get_different_ties(df.index.values, df.ids.values, df.dist.values, k)
    print(df.reindex(result))
    print(result)
    assert list(result) == [0, 1, 2, 4, 5, 6, 7, 8, 9, 11, 12, 13, 14, 15, 17, 18, 19, 20, 21, 22, 23]



from sorted_nearest.src.sorted_nearest import nearest

import pandas as pd
import numpy as np

l_s = np.array([3, 4, 7, 8])
l_e = np.array([4, 5, 9, 11])

r_s = np.array([1, 5, 8])
r_e = np.array([2, 6, 10])

idx_l, idx_r, dist = nearest(l_s, l_e, r_s, r_e)

print(idx_l, idx_r, dist)



starts = [1, 5]
import numpy as np

starts = np.array([1, 5, 13], dtype=int)
ends = np.array([3, 10, 14], dtype=int)
from sorted_nearest.src.sorted_nearest import find_clusters64

print(starts, ends)
for slack in [1, 2, 3]:
    cstarts, cends = find_clusters64(starts, ends, slack=slack)
    print(slack)
    print(cstarts, cends)



import hypothesis.strategies as st
from hypothesis import given
from hypothesis.extra.pandas import columns, data_frames, range_indexes

positions = st.integers(min_value=0, max_value=int(1e7))


def mysort(pos1, pos2):
    if pos1 > pos2:
        return pos2, pos1
    elif pos2 > pos1:
        return pos1, pos2
    else:
        return pos1, pos2 + 1


dfs = data_frames(columns=columns("Start End".split(), dtype=int), rows=st.tuples(positions, positions).map(mysort))



cimport cython
from libc.stdint cimport int32_t, int64_t

import numpy as np


def k_nearest_previous_nonoverlapping(ls, re, lidx, ridx, k, ties=None):

    # print("ls", ls)
    # print("re", re)
    # print("lidx", lidx)
    # print("ridx", ridx)

    if ls.dtype == np.int32 and not ties:
        return nearest_previous_nonoverlapping32_all(ls, re, lidx, ridx, k)
    elif ls.dtype == np.int32 and ties == "first":
        return nearest_previous_nonoverlapping32_first(ls, re, lidx, ridx, k)
    elif ls.dtype == np.int32 and ties == "last":
        return nearest_previous_nonoverlapping32_last(ls, re, lidx, ridx, k)
    elif ls.dtype == np.int32 and ties == "different":
        res = nearest_previous_nonoverlapping32_k_distances(ls, re, lidx, ridx, k)
        return res
    elif ls.dtype == np.int64 and not ties:
        return nearest_previous_nonoverlapping64_all(ls, re, lidx, ridx, k)
    elif ls.dtype == np.int64 and ties == "first":
        return nearest_previous_nonoverlapping64_first(ls, re, lidx, ridx, k)
    elif ls.dtype == np.int64 and ties == "last":
        return nearest_previous_nonoverlapping64_last(ls, re, lidx, ridx, k)
    elif ls.dtype == np.int64 and ties == "different":
        res = nearest_previous_nonoverlapping64_k_distances(ls, re, lidx, ridx, k)
        return res
    else:
        raise Exception("Starts/Ends not int64 or int32: " + str(ls.dtype))

def k_nearest_next_nonoverlapping(le, rs, lidx, ridx, k, ties=None):

    # print("dtype and ties")
    # print(le.dtype)
    # print(ties)

    if le.dtype == np.int32 and not ties:
        return nearest_next_nonoverlapping32_all(le, rs, lidx, ridx, k)
    elif le.dtype == np.int32 and ties == "first":
        return nearest_next_nonoverlapping32_first(le, rs, lidx, ridx, k)
    elif le.dtype == np.int32 and ties == "last":
        return nearest_next_nonoverlapping32_last(le, rs, lidx, ridx, k)
    elif le.dtype == np.int32 and ties == "different":
        res = nearest_next_nonoverlapping32_k_distances(le, rs, lidx, ridx, k)
        return res
    elif le.dtype == np.int64 and not ties:
        return nearest_next_nonoverlapping64_all(le, rs, lidx, ridx, k)
    elif le.dtype == np.int64 and ties == "first":
        return nearest_next_nonoverlapping64_first(le, rs, lidx, ridx, k)
    elif le.dtype == np.int64 and ties == "last":
        return nearest_next_nonoverlapping64_last(le, rs, lidx, ridx, k)
    elif le.dtype == np.int64 and ties == "different":
        res = nearest_next_nonoverlapping64_k_distances(le, rs, lidx, ridx, k)
        return res
    else:
        raise Exception("Starts/Ends not int64 or int64: " + str(le.dtype))



@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_previous_nonoverlapping32_all(const int32_t [::1] l_s, const int32_t [::1] r_e,
                                            const long [::1] l_idx, const long [::1] r_idx, const long [::1] k):

    cdef int len_l = len(l_s)
    # cdef int len_rx = len(r_idx)
    cdef int len_r = len(r_e)

    cdef int j = 0 # counter for r_e
    cdef int i = 0 # counter for l_s
    cdef int h = 0

    cdef int outarr_length = len_l
    cdef int outarr_length_old = len_l
    cdef int nfound = 0
    cdef int nfound_local = 0
    cdef int last = -1
    cdef int _dist = -1

    arr_lidx = np.ones(len_l, dtype=np.int_) * -1
    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int32) * -1

    cdef long [::1] lidx
    cdef long [::1] ridx
    cdef int32_t [::1] dist

    lidx = arr_lidx
    ridx = arr_ridx
    dist = arr_dist

    for i in range(len_l):

        # print("-----" * 10)
        # print("i", i)

        h = r_idx[i]
        nfound_local = 0
        last = r_e[h]

        while h >= 0 and nfound_local < k[i]:

            # print("h", h, "nfound_local", nfound_local)
            if nfound == outarr_length - 1:

                # print("-- resizing --" * 5)

                outarr_length_old = outarr_length
                outarr_length = (outarr_length * 2) + 1
                arr_lidx = np.resize(arr_lidx, outarr_length)
                arr_lidx[outarr_length_old:] = -1
                # print("after resizing", arr_lidx)
                lidx = arr_lidx
                arr_ridx = np.resize(arr_ridx, outarr_length)
                arr_ridx[outarr_length_old:] = -1
                ridx = arr_ridx
                arr_dist = np.resize(arr_dist, outarr_length)
                arr_dist[outarr_length_old:] = -1
                dist = arr_dist

            ridx[nfound] = h
            lidx[nfound] = l_idx[i]
            dist[nfound] = l_s[i] - r_e[h]

            # print(arr_lidx)
            # print(arr_ridx)
            # print(arr_dist)

            nfound += 1

            # print("last was", last, "r_e[h] is", r_e[h])
            if r_e[h] != last:
                nfound_local += 1

            last = r_e[h]
            h -= 1

        # print("while done. h:", h, "nfound_local:", nfound_local)

        i += 1

    return arr_lidx[:nfound], arr_ridx[:nfound], arr_dist[:nfound]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_next_nonoverlapping32_all(const int32_t [::1] l_e, const int32_t [::1] r_s,
                                        const long [::1] l_idx, const long [::1] r_idx, const long [::1] k):

    cdef int len_l = len(l_e)
    cdef int len_rx = len(r_idx)
    cdef int len_r = len(r_s)

    cdef int j = 0
    cdef int i = 0
    cdef int h = 0

    cdef int outarr_length = len_l
    cdef int outarr_length_old = len_l
    cdef int nfound = 0
    cdef int nfound_local = 0
    cdef int last = -1
    cdef int _dist = -1

    arr_lidx = np.ones(len_l, dtype=np.int_) * -1
    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int32) * -1

    cdef long [::1] lidx
    cdef long [::1] ridx
    cdef int32_t [::1] dist

    lidx = arr_lidx
    ridx = arr_ridx
    dist = arr_dist

    for i in range(len_l):

        h = r_idx[i]
        nfound_local = 0
        last = r_s[h]
        # print("-----" * 10)
        # print("i", i)

        while h < len_r and nfound_local < k[i]:

            # print("h", h, "nfound_local", nfound_local)

            if nfound == outarr_length - 1:
                outarr_length_old = outarr_length
                outarr_length = (outarr_length * 2) + 1
                arr_lidx = np.resize(arr_lidx, outarr_length)
                arr_lidx[outarr_length_old:] = -1
                lidx = arr_lidx
                arr_ridx = np.resize(arr_ridx, outarr_length)
                arr_ridx[outarr_length_old:] = -1
                ridx = arr_ridx
                arr_dist = np.resize(arr_dist, outarr_length)
                arr_dist[outarr_length_old:] = -1
                dist = arr_dist

            ridx[nfound] = h
            lidx[nfound] = l_idx[i]
            dist[nfound] = r_s[h] - l_e[i]

            # print(arr_lidx)
            # print(arr_ridx)
            # print(arr_dist)

            nfound += 1

            # print("last was", last, "r_s[h] is", r_s[h])
            if r_s[h] != last:
                nfound_local += 1

            last = r_s[h]
            h += 1

        # print("while done. h:", h, "nfound_local:", nfound_local)
        i += 1

    return arr_lidx[:nfound], arr_ridx[:nfound], arr_dist[:nfound]



@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_previous_nonoverlapping32_k_distances(const int32_t [::1] l_s, const int32_t [::1] r_e,
                                                    const long [::1] l_idx, const long [::1] r_idx, const long [::1] k):

    # print("nearest_previous_nonoverlapping32_k_distances " * 5)
    cdef int len_l = len(l_s)
    # cdef int len_rx = len(r_idx)
    cdef int len_r = len(r_e)

    cdef int j = 0 # counter for r_e
    cdef int i = 0 # counter for l_s
    cdef int h = 0

    cdef int outarr_length = len_l
    cdef int outarr_length_old = len_l
    cdef int nfound = 0
    cdef int nfound_local = 0
    cdef int last = -1
    cdef int last_dist = -1
    cdef int curr_dist = -1

    arr_lidx = np.ones(len_l, dtype=np.int_) * -1
    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int32) * -1

    cdef long [::1] lidx
    cdef long [::1] ridx
    cdef int32_t [::1] dist

    lidx = arr_lidx
    ridx = arr_ridx
    dist = arr_dist

    while i < len_l:

        h = r_idx[i]
        nfound_local = 0
        last_dist = -1

        while h >= 0 and nfound_local < k[i]:

            if nfound == outarr_length - 1:

                outarr_length_old = outarr_length
                outarr_length = (outarr_length * 2) + 1
                arr_lidx = np.resize(arr_lidx, outarr_length)
                arr_lidx[outarr_length_old:] = -1
                lidx = arr_lidx
                arr_ridx = np.resize(arr_ridx, outarr_length)
                arr_ridx[outarr_length_old:] = -1
                ridx = arr_ridx
                arr_dist = np.resize(arr_dist, outarr_length)
                arr_dist[outarr_length_old:] = -1
                dist = arr_dist

            curr_dist = l_s[i] - r_e[h]
            ridx[nfound] = h
            lidx[nfound] = l_idx[i]
            dist[nfound] = curr_dist

            nfound += 1

            if curr_dist != last_dist:
                nfound_local += 1

            curr_dist = last_dist

            last = r_e[h]
            h -= 1

        i += 1

    return arr_lidx[:nfound], arr_ridx[:nfound], arr_dist[:nfound]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_next_nonoverlapping32_k_distances(const int32_t [::1] l_e, const int32_t [::1] r_s,
                                        const long [::1] l_idx, const long [::1] r_idx, const long [::1] k):

    cdef int len_l = len(l_e)
    cdef int len_rx = len(r_idx)
    cdef int len_r = len(r_s)

    cdef int j = 0
    cdef int i = 0
    cdef int h = 0

    cdef int outarr_length = len_l
    cdef int outarr_length_old = len_l
    cdef int nfound = 0
    cdef int nfound_local = 0
    cdef int last = -1
    cdef int curr_dist = -1
    cdef int last_dist = -1


    arr_lidx = np.ones(len_l, dtype=np.int_) * -1
    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int32) * -1

    cdef long [::1] lidx
    cdef long [::1] ridx
    cdef int32_t [::1] dist

    lidx = arr_lidx
    ridx = arr_ridx
    dist = arr_dist

    while i < len_l:

        h = r_idx[i]
        nfound_local = 0
        last = -1

        while h < len_r and nfound_local < k[i]:

            if nfound == outarr_length - 1:

                outarr_length_old = outarr_length
                outarr_length = (outarr_length * 2) + 1
                arr_lidx = np.resize(arr_lidx, outarr_length)
                arr_lidx[outarr_length_old:] = -1
                lidx = arr_lidx
                arr_ridx = np.resize(arr_ridx, outarr_length)
                arr_ridx[outarr_length_old:] = -1
                ridx = arr_ridx
                arr_dist = np.resize(arr_dist, outarr_length)
                arr_dist[outarr_length_old:] = -1
                dist = arr_dist


            curr_dist = r_s[h] - l_e[i]

            ridx[nfound] = h
            lidx[nfound] = l_idx[i]
            dist[nfound] = curr_dist

            nfound += 1

            if curr_dist != last_dist:
                nfound_local += 1

            last_dist = curr_dist

            last = r_s[h]
            h += 1

        i += 1

    return arr_lidx[:nfound], arr_ridx[:nfound], arr_dist[:nfound]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_next_nonoverlapping32_first(const int32_t [::1] l_e, const int32_t [::1] r_s,
                                        const long [::1] l_idx, const long [::1] r_idx, const long [::1] k):

    # print("nearest_next_nonoverlapping first")

    cdef int len_l = len(l_e)
    cdef int len_rx = len(r_idx)
    cdef int len_r = len(r_s)

    cdef int j = 0
    cdef int i = 0
    cdef int h = 0

    cdef int outarr_length = len_l
    cdef int outarr_length_old = len_l
    cdef int nfound = 0
    cdef int nfound_local = 0
    cdef int last = -1
    cdef int curr_dist = -1
    cdef int last_dist = -1


    arr_lidx = np.ones(len_l, dtype=np.int_) * -1
    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int32) * -1

    cdef long [::1] lidx
    cdef long [::1] ridx
    cdef int32_t [::1] dist

    lidx = arr_lidx
    ridx = arr_ridx
    dist = arr_dist

    # print("ridx", list(r_idx))
    # print("lidx", list(l_idx))
    # print("l_e", list(l_e))
    # print("r_s", list(r_s))
    # print("i", i)
    # print("len_l", len_l)
    # print("len_rx", len_rx)
    # print("j", j)
    while i < len_l:

        nfound_local = 0
        last_dist = -1

        h = r_idx[i]

        while h < len_r and nfound_local < k[i]:

            # print("h", h)
            # print("rs", r_s[h])
            # print("le", l_e[i])

            if nfound >= outarr_length - 1:


                outarr_length_old = outarr_length
                outarr_length = (outarr_length * 2) + 1
                arr_lidx = np.resize(arr_lidx, outarr_length)
                arr_lidx[outarr_length_old:] = -1
                lidx = arr_lidx
                arr_ridx = np.resize(arr_ridx, outarr_length)
                arr_ridx[outarr_length_old:] = -1
                ridx = arr_ridx
                arr_dist = np.resize(arr_dist, outarr_length)
                arr_dist[outarr_length_old:] = -1
                dist = arr_dist

            curr_dist = r_s[h] - l_e[i]

            if curr_dist != last_dist:
                # print("inserting ridx", h, "lidx", l_idx[i], "dist", curr_dist)
                ridx[nfound] = h
                lidx[nfound] = l_idx[i]
                dist[nfound] = curr_dist

                nfound += 1

                nfound_local += 1

                last_dist = curr_dist

            last = r_s[h]
            h += 1

        i += 1

    return arr_lidx[:nfound], arr_ridx[:nfound], arr_dist[:nfound]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_previous_nonoverlapping32_first(const int32_t [::1] l_s, const int32_t [::1] r_e,
                                              const long [::1] l_idx, const long [::1] r_idx,
                                              const long [::1] k):

    # print("nearest previous nonoverlapping first")

    cdef int len_l = len(l_s)
    cdef int len_rx = len(r_idx)
    cdef int len_r = len(r_e)

    cdef int j = 0 # counter for r_e
    cdef int i = 0 # counter for l_s
    cdef int h = 0

    cdef int outarr_length = len_l
    cdef int outarr_length_old = len_l
    cdef int nfound = 0
    cdef int nfound_local = 0
    cdef int last = -1
    cdef int last_dist = -1
    cdef int curr_dist = -1

    arr_lidx = np.ones(len_l, dtype=np.int_) * -1
    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int32) * -1

    cdef long [::1] lidx
    cdef long [::1] ridx
    cdef int32_t [::1] dist

    lidx = arr_lidx
    ridx = arr_ridx
    dist = arr_dist

    # print("r_idx", list(r_idx))
    # print("r_e", list(r_e))
    # print("k", list(k))

    while i < len_l:

        last_dist = -1
        nfound_local = 0
        h = r_idx[i]

        while nfound_local < k[i] and h >= 0:  # and l_s[i] - r_e[h] >= 0:

            # print("h", h)
            # print("ls", l_s[i])
            # print("re", r_e[h])

            if nfound >= outarr_length - 1:
                outarr_length_old = outarr_length
                outarr_length = (outarr_length * 2) + 1
                arr_lidx = np.resize(arr_lidx, outarr_length)
                arr_lidx[outarr_length_old:] = -1
                lidx = arr_lidx
                arr_ridx = np.resize(arr_ridx, outarr_length)
                arr_ridx[outarr_length_old:] = -1
                ridx = arr_ridx
                arr_dist = np.resize(arr_dist, outarr_length)
                arr_dist[outarr_length_old:] = -1
                dist = arr_dist

            curr_dist = l_s[i] - r_e[h]

            if curr_dist != last_dist:
                # print("inserting ridx", h, "lidx", l_idx[i], "dist", curr_dist)
                ridx[nfound] = h
                lidx[nfound] = l_idx[i]
                dist[nfound] = curr_dist
                last_dist = curr_dist

                nfound += 1

                nfound_local += 1

            last = r_e[h]
            h -= 1
            # print("h", h, "len_rx", len_rx)

        i += 1
        # print("i", i, "len_l", len_l)

    return arr_lidx[:nfound], arr_ridx[:nfound], arr_dist[:nfound]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_next_nonoverlapping32_last(const int32_t [::1] l_e, const int32_t [::1] r_s,
                                        const long [::1] l_idx, const long [::1] r_idx, const long [::1] k):

    # print("  nearest_next_nonoverlapping last" * 5)

    cdef int len_l = len(l_e)
    cdef int len_rx = len(r_idx)
    cdef int len_r = len(r_s)

    cdef int j = 0
    cdef int i = 0
    cdef int h = 0

    cdef int outarr_length = len_l
    cdef int outarr_length_old = 0
    cdef int nfound = 0
    cdef int nfound_local = 0
    cdef int last = -1
    cdef int curr_dist = -1
    cdef int last_dist = -1
    cdef int start_dist = -1


    arr_lidx = np.ones(len_l, dtype=np.int_) * -1
    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int64) * -1

    cdef long [::1] lidx
    cdef long [::1] ridx
    cdef long [::1] dist

    lidx = arr_lidx
    ridx = arr_ridx
    dist = arr_dist

    for i in range(len_l):

        # print("i", i)
        nfound_local = 0
        h = r_idx[i]
        last_dist = r_s[h] - l_e[i]
        start_dist = last_dist

        while h < len_r and nfound_local < k[i]:

            # print("h", h)
            # print("rs", r_s[h])
            # print("le", l_e[i])

            if nfound >= outarr_length - 1:

                # print("--resizing--" * 5)
                # print("before resizing", arr_lidx)

                outarr_length_old = outarr_length
                outarr_length = (outarr_length * 2) + 1
                arr_lidx = np.resize(arr_lidx, outarr_length)
                arr_lidx[outarr_length_old:] = -1
                # print("after resizing", arr_lidx)
                lidx = arr_lidx
                arr_ridx = np.resize(arr_ridx, outarr_length)
                arr_ridx[outarr_length_old:] = -1
                ridx = arr_ridx
                arr_dist = np.resize(arr_dist, outarr_length)
                arr_dist[outarr_length_old:] = -1
                dist = arr_dist

            curr_dist = r_s[h] - l_e[i]

            """Cases:
1. all dist per i equal. need to add last dist manually
2. dists unequal -> all will be added but last. Need to add last manually
"""
            # if unequal, begin searching for last for next distance
            if curr_dist != last_dist:
                # print("if unequal, begin searching for last for next distance")
                nfound += 1
                nfound_local += 1

            # for each entry, add distance
            ridx[nfound] = h
            lidx[nfound] = l_idx[i]
            dist[nfound] = curr_dist


            # print("lidx", list(lidx))
            # print("ridx", list(ridx))
            # print("dist", list(dist))

            last_dist = curr_dist

            h += 1

        # in case of all equal, need to add manually
        if start_dist == last_dist and k[i]:
            # print("in case of all equal, need to add manually")
            ridx[nfound] = h - 1
            lidx[nfound] = l_idx[i]
            dist[nfound] = last_dist
            # print("lidx", list(lidx))
            # print("ridx", list(ridx))
            # print("dist", list(dist))
            nfound += 1


        # in case last dist was not added
        elif nfound_local < k[i] and dist[nfound - 1] != last_dist:
            # print("FINALLY inserting lidx", l_idx[i], "ridx", h - 1, "dist", curr_dist)

            ridx[nfound] = h - 1
            lidx[nfound] = l_idx[i]
            dist[nfound] = last_dist
            nfound += 1
            # print("lidx", list(lidx))
            # print("ridx", list(ridx))
            # print("dist", list(dist))

        i += 1

    return arr_lidx[:nfound], arr_ridx[:nfound], arr_dist[:nfound]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_previous_nonoverlapping32_last(const int32_t [::1] l_s, const int32_t [::1] r_e,
                                              const long [::1] l_idx, const long [::1] r_idx,
                                              const long [::1] k):

    # print("  nearest previous nonoverlapping last" * 5)

    cdef int len_l = len(l_s)
    cdef int len_rx = len(r_idx)
    cdef int len_r = len(r_e)

    cdef int i = 0 # counter for r_e
    cdef int h = 0

    cdef int outarr_length = len_l
    cdef int outarr_length_old = -1
    cdef int nfound = 0
    cdef int nfound_local = 0
    cdef int last = -1
    cdef int last_dist = -1
    cdef int curr_dist = -1

    arr_lidx = np.ones(len_l, dtype=np.int_) * -1
    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int64) * -1

    cdef long [::1] lidx
    cdef long [::1] ridx
    cdef long [::1] dist

    lidx = arr_lidx
    ridx = arr_ridx
    dist = arr_dist

    # print("l_idx", list(l_idx))
    # print("r_idx", list(r_idx))
    # print("r_e", list(r_e))
    # print("k", list(k))
    # print("len_l", len_l)

    for i in range(len_l):

        nfound_local = 0
        h = r_idx[i]
        last_dist = -1

        # print("----" * 5)
        # print("i", i)

        while nfound_local < k[i] and h >= 0:  # and l_s[i] - r_e[h] >= 0:

            # print("***" * 3)
            # print("h", h)
            # print("ls", l_s[i])
            # print("re", r_e[h])

            if nfound >= outarr_length - 1:

                outarr_length_old = outarr_length
                outarr_length = (outarr_length * 2) + 1
                arr_lidx = np.resize(arr_lidx, outarr_length)
                arr_lidx[outarr_length_old:] = -1
                # print("after resizing", arr_lidx)
                lidx = arr_lidx
                arr_ridx = np.resize(arr_ridx, outarr_length)
                arr_ridx[outarr_length_old:] = -1
                ridx = arr_ridx
                arr_dist = np.resize(arr_dist, outarr_length)
                arr_dist[outarr_length_old:] = -1
                dist = arr_dist

            curr_dist = l_s[i] - r_e[h]

            # print("inserting lidx", l_idx[i], "ridx", h, "dist", curr_dist, "i", i)
            ridx[nfound] = h
            lidx[nfound] = l_idx[i]
            dist[nfound] = curr_dist

            # print("curr_dist", curr_dist, "vs last_dist", last_dist)
            if curr_dist != last_dist:
                nfound += 1
                nfound_local += 1
                # print("ridx", list(ridx))
                # print("nfound", nfound)
            last_dist = curr_dist

            # print("lidx", list(lidx))
            # print("ridx", list(ridx))
            # print("dist", list(dist))

            h -= 1

        if nfound > 0 and nfound_local < k[i] and dist[nfound - 1] != curr_dist:

            # print("FINALLY inserting lidx", l_idx[i], "ridx", h + 1, "dist", curr_dist)
            ridx[nfound] = h + 1
            lidx[nfound] = l_idx[i]
            dist[nfound] = curr_dist
            nfound += 1

            # print("lidx", list(lidx)[:nfound])
            # print("ridx", list(ridx)[:nfound])
            # print("dist", list(dist)[:nfound])

        i += 1
        # print("i", i, "len_l", len_l)

    return arr_lidx[:nfound], arr_ridx[:nfound], arr_dist[:nfound]




@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef get_all_ties(const int64_t[::1] lx, const int64_t[::1] ids, const int64_t[::1] dist, int k):

    # print("lx", list(lx))
    # print("ids", list(ids))
    # print("dist", list(dist))

    # for each index
    # want to keep until we have k
    # then keep all with same distance
    cdef:
        int i = 0
        int first_i = 0
        int current_found = 0
        int nfound = 0
        int current_id = ids[0]
        int last_id = -1
        int length = len(lx)
        int current_dist = -1
        int last_dist = -1

    arr_lidx = np.ones(length, dtype=np.int_) * -1

    cdef long [::1] lidx

    lidx = arr_lidx

    # Chromosome  Start  End Name  Score Strand  __k__  __IX__  Start_b  End_b Name_b  Score_b Strand_b  Distance
    # 0       chr1      1    2    a      0      -      2       0        1      2      a        0        -         0
    # 1       chr1      1    2    a      0      -      2       0        1      2      a        0        -         0
    # 2       chr1      1    2    a      0      -      2       0        1      2      a        0        -         0

    while i < length:

        # print("i", i)
        # print("ids[i]", ids[i])
        # print("last_id", last_id)
        # print("current_found", current_found)
        # print("nfound", nfound)
        # print("length", length)
        first_i = i

        if ids[i] != last_id:
            # we have come to a new id, need to start counting again


            current_found = 0
            current_id = ids[i]
            last_id = ids[i]
            continue

        # while have not found k different ties, continue
        while i < length and current_found < k:
            # print("**first while**")
            # print("i", i, "length", length, "ids[i]", ids[i], "last_id", last_id, "dist[i]", last_dist, "current_found", current_found)
            lidx[nfound] = lx[i]
            # print("lidx", arr_lidx)
            last_dist = dist[i]
            nfound += 1
            current_found += 1
            i += 1

        # while we have not found all ties with same length as last included
        while i < length and ids[i] == last_id and dist[i] == last_dist:
            # print("**second while**")
            # print("i", i, "length", length, "ids[i]", ids[i], "last_id", last_id, "dist[i]", last_dist, "current_found", current_found)
            lidx[nfound] = lx[i]
            # print("lidx", arr_lidx)
            nfound += 1
            i += 1

        if i == first_i:
            i += 1

        # print(arr_lidx)

    return arr_lidx[:nfound]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_previous_nonoverlapping64_all(const int64_t [::1] l_s, const int64_t [::1] r_e,
                                            const long [::1] l_idx, const long [::1] r_idx, const long [::1] k):

    cdef int len_l = len(l_s)
    # cdef int len_rx = len(r_idx)
    cdef int len_r = len(r_e)

    cdef int j = 0 # counter for r_e
    cdef int i = 0 # counter for l_s
    cdef int h = 0

    cdef int outarr_length = len_l
    cdef int outarr_length_old = len_l
    cdef int nfound = 0
    cdef int nfound_local = 0
    cdef int last = -1
    cdef int _dist = -1

    arr_lidx = np.ones(len_l, dtype=np.int_) * -1
    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int64) * -1

    cdef long [::1] lidx
    cdef long [::1] ridx
    cdef int64_t [::1] dist

    lidx = arr_lidx
    ridx = arr_ridx
    dist = arr_dist

    for i in range(len_l):

        # print("-----" * 10)
        # print("i", i)

        h = r_idx[i]
        nfound_local = 0
        last = r_e[h]

        while h >= 0 and nfound_local < k[i]:

            # print("h", h, "nfound_local", nfound_local)
            if nfound == outarr_length - 1:

                # print("-- resizing --" * 5)

                outarr_length_old = outarr_length
                outarr_length = (outarr_length * 2) + 1
                arr_lidx = np.resize(arr_lidx, outarr_length)
                arr_lidx[outarr_length_old:] = -1
                # print("after resizing", arr_lidx)
                lidx = arr_lidx
                arr_ridx = np.resize(arr_ridx, outarr_length)
                arr_ridx[outarr_length_old:] = -1
                ridx = arr_ridx
                arr_dist = np.resize(arr_dist, outarr_length)
                arr_dist[outarr_length_old:] = -1
                dist = arr_dist

            ridx[nfound] = h
            lidx[nfound] = l_idx[i]
            dist[nfound] = l_s[i] - r_e[h]

            # print(arr_lidx)
            # print(arr_ridx)
            # print(arr_dist)

            nfound += 1

            # print("last was", last, "r_e[h] is", r_e[h])
            if r_e[h] != last:
                nfound_local += 1

            last = r_e[h]
            h -= 1

        # print("while done. h:", h, "nfound_local:", nfound_local)

        i += 1

    return arr_lidx[:nfound], arr_ridx[:nfound], arr_dist[:nfound]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_next_nonoverlapping64_all(const int64_t [::1] l_e, const int64_t [::1] r_s,
                                        const long [::1] l_idx, const long [::1] r_idx, const long [::1] k):

    cdef int len_l = len(l_e)
    cdef int len_rx = len(r_idx)
    cdef int len_r = len(r_s)

    cdef int j = 0
    cdef int i = 0
    cdef int h = 0

    cdef int outarr_length = len_l
    cdef int outarr_length_old = len_l
    cdef int nfound = 0
    cdef int nfound_local = 0
    cdef int last = -1
    cdef int _dist = -1

    arr_lidx = np.ones(len_l, dtype=np.int_) * -1
    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int64) * -1

    cdef long [::1] lidx
    cdef long [::1] ridx
    cdef int64_t [::1] dist

    lidx = arr_lidx
    ridx = arr_ridx
    dist = arr_dist

    for i in range(len_l):

        h = r_idx[i]
        nfound_local = 0
        last = r_s[h]
        # print("-----" * 10)
        # print("i", i)

        while h < len_r and nfound_local < k[i]:

            # print("h", h, "nfound_local", nfound_local)

            if nfound == outarr_length - 1:
                outarr_length_old = outarr_length
                outarr_length = (outarr_length * 2) + 1
                arr_lidx = np.resize(arr_lidx, outarr_length)
                arr_lidx[outarr_length_old:] = -1
                lidx = arr_lidx
                arr_ridx = np.resize(arr_ridx, outarr_length)
                arr_ridx[outarr_length_old:] = -1
                ridx = arr_ridx
                arr_dist = np.resize(arr_dist, outarr_length)
                arr_dist[outarr_length_old:] = -1
                dist = arr_dist

            ridx[nfound] = h
            lidx[nfound] = l_idx[i]
            dist[nfound] = r_s[h] - l_e[i]

            # print(arr_lidx)
            # print(arr_ridx)
            # print(arr_dist)

            nfound += 1

            # print("last was", last, "r_s[h] is", r_s[h])
            if r_s[h] != last:
                nfound_local += 1

            last = r_s[h]
            h += 1

        # print("while done. h:", h, "nfound_local:", nfound_local)
        i += 1

    return arr_lidx[:nfound], arr_ridx[:nfound], arr_dist[:nfound]



@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_previous_nonoverlapping64_k_distances(const int64_t [::1] l_s, const int64_t [::1] r_e,
                                                    const long [::1] l_idx, const long [::1] r_idx, const long [::1] k):

    cdef int len_l = len(l_s)
    # cdef int len_rx = len(r_idx)
    cdef int len_r = len(r_e)

    cdef int j = 0 # counter for r_e
    cdef int i = 0 # counter for l_s
    cdef int h = 0

    cdef int outarr_length = len_l
    cdef int outarr_length_old = len_l
    cdef int nfound = 0
    cdef int nfound_local = 0
    cdef int last = -1
    cdef int last_dist = -1
    cdef int curr_dist = -1

    arr_lidx = np.ones(len_l, dtype=np.int_) * -1
    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int64) * -1

    cdef long [::1] lidx
    cdef long [::1] ridx
    cdef int64_t [::1] dist

    lidx = arr_lidx
    ridx = arr_ridx
    dist = arr_dist

    while i < len_l and j < len_r:

        h = r_idx[j]
        nfound_local = 0
        j += 1
        last_dist = -1

        while h >= 0 and nfound_local < k[i]:

            if nfound == outarr_length - 1:

                outarr_length_old = outarr_length
                outarr_length = (outarr_length * 2) + 1
                arr_lidx = np.resize(arr_lidx, outarr_length)
                arr_lidx[outarr_length_old:] = -1
                lidx = arr_lidx
                arr_ridx = np.resize(arr_ridx, outarr_length)
                arr_ridx[outarr_length_old:] = -1
                ridx = arr_ridx
                arr_dist = np.resize(arr_dist, outarr_length)
                arr_dist[outarr_length_old:] = -1
                dist = arr_dist

            curr_dist = l_s[i] - r_e[h]
            ridx[nfound] = h
            lidx[nfound] = l_idx[i]
            dist[nfound] = curr_dist

            nfound += 1

            if curr_dist != last_dist:
                nfound_local += 1

            curr_dist = last_dist

            last = r_e[h]
            h -= 1

        i += 1

    return arr_lidx[:nfound], arr_ridx[:nfound], arr_dist[:nfound]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_next_nonoverlapping64_k_distances(const int64_t [::1] l_e, const int64_t [::1] r_s,
                                        const long [::1] l_idx, const long [::1] r_idx, const long [::1] k):

    cdef int len_l = len(l_e)
    cdef int len_rx = len(r_idx)
    cdef int len_r = len(r_s)

    cdef int j = 0
    cdef int i = 0
    cdef int h = 0

    cdef int outarr_length = len_l
    cdef int outarr_length_old = len_l
    cdef int nfound = 0
    cdef int nfound_local = 0
    cdef int last = -1
    cdef int curr_dist = -1
    cdef int last_dist = -1


    arr_lidx = np.ones(len_l, dtype=np.int_) * -1
    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int64) * -1

    cdef long [::1] lidx
    cdef long [::1] ridx
    cdef int64_t [::1] dist

    lidx = arr_lidx
    ridx = arr_ridx
    dist = arr_dist

    while i < len_l and j < len_rx:

        h = r_idx[j]
        nfound_local = 0
        j += 1
        last = -1

        while h < len_r and nfound_local < k[i]:

            if nfound == outarr_length - 1:

                outarr_length_old = outarr_length
                outarr_length = (outarr_length * 2) + 1
                arr_lidx = np.resize(arr_lidx, outarr_length)
                arr_lidx[outarr_length_old:] = -1
                lidx = arr_lidx
                arr_ridx = np.resize(arr_ridx, outarr_length)
                arr_ridx[outarr_length_old:] = -1
                ridx = arr_ridx
                arr_dist = np.resize(arr_dist, outarr_length)
                arr_dist[outarr_length_old:] = -1
                dist = arr_dist

            curr_dist = r_s[h] - l_e[i]

            ridx[nfound] = h
            lidx[nfound] = l_idx[i]
            dist[nfound] = curr_dist

            nfound += 1

            if curr_dist != last_dist:
                nfound_local += 1

            last_dist = curr_dist

            last = r_s[h]
            h += 1

        i += 1

    return arr_lidx[:nfound], arr_ridx[:nfound], arr_dist[:nfound]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_next_nonoverlapping64_first(const int64_t [::1] l_e, const int64_t [::1] r_s,
                                        const long [::1] l_idx, const long [::1] r_idx, const long [::1] k):

    # print("nearest_next_nonoverlapping first")

    cdef int len_l = len(l_e)
    cdef int len_rx = len(r_idx)
    cdef int len_r = len(r_s)

    cdef int j = 0
    cdef int i = 0
    cdef int h = 0

    cdef int outarr_length = len_l
    cdef int outarr_length_old = len_l
    cdef int nfound = 0
    cdef int nfound_local = 0
    cdef int last = -1
    cdef int curr_dist = -1
    cdef int last_dist = -1


    arr_lidx = np.ones(len_l, dtype=np.int_) * -1
    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int64) * -1

    cdef long [::1] lidx
    cdef long [::1] ridx
    cdef int64_t [::1] dist

    lidx = arr_lidx
    ridx = arr_ridx
    dist = arr_dist

    # print("ridx", list(r_idx))
    # print("lidx", list(l_idx))
    # print("l_e", list(l_e))
    # print("r_s", list(r_s))
    # print("i", i)
    # print("len_l", len_l)
    # print("len_rx", len_rx)
    # print("j", j)
    while i < len_l:

        nfound_local = 0
        last_dist = -1

        h = r_idx[i]

        while h < len_r and nfound_local < k[i]:

            # print("h", h)
            # print("rs", r_s[h])
            # print("le", l_e[i])

            if nfound >= outarr_length - 1:


                outarr_length_old = outarr_length
                outarr_length = (outarr_length * 2) + 1
                arr_lidx = np.resize(arr_lidx, outarr_length)
                arr_lidx[outarr_length_old:] = -1
                lidx = arr_lidx
                arr_ridx = np.resize(arr_ridx, outarr_length)
                arr_ridx[outarr_length_old:] = -1
                ridx = arr_ridx
                arr_dist = np.resize(arr_dist, outarr_length)
                arr_dist[outarr_length_old:] = -1
                dist = arr_dist

            curr_dist = r_s[h] - l_e[i]

            if curr_dist != last_dist:
                # print("inserting ridx", h, "lidx", l_idx[i], "dist", curr_dist)
                ridx[nfound] = h
                lidx[nfound] = l_idx[i]
                dist[nfound] = curr_dist

                nfound += 1

                nfound_local += 1

                last_dist = curr_dist

            last = r_s[h]
            h += 1

        i += 1

    return arr_lidx[:nfound], arr_ridx[:nfound], arr_dist[:nfound]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_previous_nonoverlapping64_first(const int64_t [::1] l_s, const int64_t [::1] r_e,
                                              const long [::1] l_idx, const long [::1] r_idx,
                                              const long [::1] k):

    # print("nearest previous nonoverlapping first")

    cdef int len_l = len(l_s)
    cdef int len_rx = len(r_idx)
    cdef int len_r = len(r_e)

    cdef int j = 0 # counter for r_e
    cdef int i = 0 # counter for l_s
    cdef int h = 0

    cdef int outarr_length = len_l
    cdef int outarr_length_old = len_l
    cdef int nfound = 0
    cdef int nfound_local = 0
    cdef int last = -1
    cdef int last_dist = -1
    cdef int curr_dist = -1

    arr_lidx = np.ones(len_l, dtype=np.int_) * -1
    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int64) * -1

    cdef long [::1] lidx
    cdef long [::1] ridx
    cdef int64_t [::1] dist

    lidx = arr_lidx
    ridx = arr_ridx
    dist = arr_dist

    # print("r_idx", list(r_idx))
    # print("r_e", list(r_e))
    # print("k", list(k))

    while i < len_l:

        last_dist = -1
        nfound_local = 0
        h = r_idx[i]

        while nfound_local < k[i] and h >= 0:  # and l_s[i] - r_e[h] >= 0:

            # print("h", h)
            # print("ls", l_s[i])
            # print("re", r_e[h])

            if nfound >= outarr_length - 1:
                outarr_length_old = outarr_length
                outarr_length = (outarr_length * 2) + 1
                arr_lidx = np.resize(arr_lidx, outarr_length)
                arr_lidx[outarr_length_old:] = -1
                lidx = arr_lidx
                arr_ridx = np.resize(arr_ridx, outarr_length)
                arr_ridx[outarr_length_old:] = -1
                ridx = arr_ridx
                arr_dist = np.resize(arr_dist, outarr_length)
                arr_dist[outarr_length_old:] = -1
                dist = arr_dist

            curr_dist = l_s[i] - r_e[h]

            if curr_dist != last_dist:
                # print("inserting ridx", h, "lidx", l_idx[i], "dist", curr_dist)
                ridx[nfound] = h
                lidx[nfound] = l_idx[i]
                dist[nfound] = curr_dist
                last_dist = curr_dist

                nfound += 1

                nfound_local += 1

            last = r_e[h]
            h -= 1
            # print("h", h, "len_rx", len_rx)

        i += 1
        # print("i", i, "len_l", len_l)

    return arr_lidx[:nfound], arr_ridx[:nfound], arr_dist[:nfound]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_next_nonoverlapping64_last(const int64_t [::1] l_e, const int64_t [::1] r_s,
                                        const long [::1] l_idx, const long [::1] r_idx, const long [::1] k):

    # print("  nearest_next_nonoverlapping last" * 5)

    cdef int len_l = len(l_e)
    cdef int len_rx = len(r_idx)
    cdef int len_r = len(r_s)

    cdef int j = 0
    cdef int i = 0
    cdef int h = 0

    cdef int outarr_length = len_l
    cdef int outarr_length_old = 0
    cdef int nfound = 0
    cdef int nfound_local = 0
    cdef int last = -1
    cdef int curr_dist = -1
    cdef int last_dist = -1
    cdef int start_dist = -1


    arr_lidx = np.ones(len_l, dtype=np.int_) * -1
    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int64) * -1

    cdef long [::1] lidx
    cdef long [::1] ridx
    cdef long [::1] dist

    lidx = arr_lidx
    ridx = arr_ridx
    dist = arr_dist

    for i in range(len_l):

        # print("i", i)
        nfound_local = 0
        h = r_idx[i]
        last_dist = r_s[h] - l_e[i]
        start_dist = last_dist

        while h < len_r and nfound_local < k[i]:

            # print("h", h)
            # print("rs", r_s[h])
            # print("le", l_e[i])

            if nfound >= outarr_length - 1:

                # print("--resizing--" * 5)
                # print("before resizing", arr_lidx)

                outarr_length_old = outarr_length
                outarr_length = (outarr_length * 2) + 1
                arr_lidx = np.resize(arr_lidx, outarr_length)
                arr_lidx[outarr_length_old:] = -1
                # print("after resizing", arr_lidx)
                lidx = arr_lidx
                arr_ridx = np.resize(arr_ridx, outarr_length)
                arr_ridx[outarr_length_old:] = -1
                ridx = arr_ridx
                arr_dist = np.resize(arr_dist, outarr_length)
                arr_dist[outarr_length_old:] = -1
                dist = arr_dist

            curr_dist = r_s[h] - l_e[i]

            """Cases:
1. all dist per i equal. need to add last dist manually
2. dists unequal -> all will be added but last. Need to add last manually
"""
            # if unequal, begin searching for last for next distance
            if curr_dist != last_dist:
                # print("if unequal, begin searching for last for next distance")
                nfound += 1
                nfound_local += 1

            # for each entry, add distance
            ridx[nfound] = h
            lidx[nfound] = l_idx[i]
            dist[nfound] = curr_dist


            # print("lidx", list(lidx))
            # print("ridx", list(ridx))
            # print("dist", list(dist))

            last_dist = curr_dist

            h += 1

        # in case of all equal, need to add manually
        if start_dist == last_dist and k[i]:
            # print("in case of all equal, need to add manually")
            ridx[nfound] = h - 1
            lidx[nfound] = l_idx[i]
            dist[nfound] = last_dist
            # print("lidx", list(lidx))
            # print("ridx", list(ridx))
            # print("dist", list(dist))
            nfound += 1


        # in case last dist was not added
        elif nfound_local < k[i] and dist[nfound - 1] != last_dist:
            # print("FINALLY inserting lidx", l_idx[i], "ridx", h - 1, "dist", curr_dist)

            ridx[nfound] = h - 1
            lidx[nfound] = l_idx[i]
            dist[nfound] = last_dist
            nfound += 1
            # print("lidx", list(lidx))
            # print("ridx", list(ridx))
            # print("dist", list(dist))

        i += 1

    return arr_lidx[:nfound], arr_ridx[:nfound], arr_dist[:nfound]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef nearest_previous_nonoverlapping64_last(const int64_t [::1] l_s, const int64_t [::1] r_e,
                                              const long [::1] l_idx, const long [::1] r_idx,
                                              const long [::1] k):

    # print("  nearest previous nonoverlapping last" * 5)

    cdef int len_l = len(l_s)
    cdef int len_rx = len(r_idx)
    cdef int len_r = len(r_e)

    cdef int i = 0 # counter for r_e
    cdef int h = 0

    cdef int outarr_length = len_l
    cdef int outarr_length_old = -1
    cdef int nfound = 0
    cdef int nfound_local = 0
    cdef int last = -1
    cdef int last_dist = -1
    cdef int curr_dist = -1

    arr_lidx = np.ones(len_l, dtype=np.int_) * -1
    arr_ridx = np.ones(len_l, dtype=np.int_) * -1
    arr_dist = np.ones(len_l, dtype=np.int64) * -1

    cdef long [::1] lidx
    cdef long [::1] ridx
    cdef long [::1] dist

    lidx = arr_lidx
    ridx = arr_ridx
    dist = arr_dist

    # print("l_idx", list(l_idx))
    # print("r_idx", list(r_idx))
    # print("r_e", list(r_e))
    # print("k", list(k))
    # print("len_l", len_l)

    for i in range(len_l):

        nfound_local = 0
        h = r_idx[i]
        last_dist = -1

        # print("----" * 5)
        # print("i", i)

        while nfound_local < k[i] and h >= 0:  # and l_s[i] - r_e[h] >= 0:

            # print("***" * 3)
            # print("h", h)
            # print("ls", l_s[i])
            # print("re", r_e[h])

            if nfound >= outarr_length - 1:

                outarr_length_old = outarr_length
                outarr_length = (outarr_length * 2) + 1
                arr_lidx = np.resize(arr_lidx, outarr_length)
                arr_lidx[outarr_length_old:] = -1
                # print("after resizing", arr_lidx)
                lidx = arr_lidx
                arr_ridx = np.resize(arr_ridx, outarr_length)
                arr_ridx[outarr_length_old:] = -1
                ridx = arr_ridx
                arr_dist = np.resize(arr_dist, outarr_length)
                arr_dist[outarr_length_old:] = -1
                dist = arr_dist

            curr_dist = l_s[i] - r_e[h]

            # print("inserting lidx", l_idx[i], "ridx", h, "dist", curr_dist, "i", i)
            ridx[nfound] = h
            lidx[nfound] = l_idx[i]
            dist[nfound] = curr_dist

            # print("curr_dist", curr_dist, "vs last_dist", last_dist)
            if curr_dist != last_dist:
                nfound += 1
                nfound_local += 1
                # print("ridx", list(ridx))
                # print("nfound", nfound)
            last_dist = curr_dist

            # print("lidx", list(lidx))
            # print("ridx", list(ridx))
            # print("dist", list(dist))

            h -= 1

        if nfound > 0 and nfound_local < k[i] and dist[nfound - 1] != curr_dist:

            # print("FINALLY inserting lidx", l_idx[i], "ridx", h + 1, "dist", curr_dist)
            ridx[nfound] = h + 1
            lidx[nfound] = l_idx[i]
            dist[nfound] = curr_dist
            nfound += 1

            # print("lidx", list(lidx)[:nfound])
            # print("ridx", list(ridx)[:nfound])
            # print("dist", list(dist)[:nfound])

        i += 1
        # print("i", i, "len_l", len_l)

    return arr_lidx[:nfound], arr_ridx[:nfound], arr_dist[:nfound]



#!/usr/bin/env python

cimport cython
from libc.stdint cimport int32_t

import numpy as np


def max_disjoint(indexes, starts, ends, slack):

    if starts.dtype == np.int_:
        return max_disjoint64(indexes, starts, ends, slack)
    elif starts.dtype == np.int32:
        return max_disjoint32(indexes, starts, ends, slack)
    else:
        raise Exception("Starts/Ends not int64 or int32: " + str(starts.dtype))


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef max_disjoint64(const long [::1] indexes, const long [::1] starts, const long [::1] ends, int slack):

    cdef int length = len(starts)
    cdef int count = 0
    cdef int r1 = 0
    cdef int l1 = 0
    cdef int r2 = 0

    output_arr_indexes = np.ones(length, dtype=np.int_) * -1
    # output_arr_start = np.ones(length, dtype=np.int_) * -1
    # output_arr_end = np.zeros(length, dtype=np.int_) * -1

    cdef long [::1] output_indexes
    # cdef long [::1] output_start
    # cdef long [::1] output_end

    output_indexes = output_arr_indexes
    # output_start = output_arr_start
    # output_end = output_arr_end

    # output_start[count] = starts[count]
    # output_end[count] = ends[count]
    output_indexes[count] = indexes[count]
    count += 1

    r1 = ends[0]
    for i in range(1, length):
        l1 = starts[i]
        r2 = ends[i]

        if l1 > (r1 + slack):
            # output_start[count] = l1
            # output_end[count] = r2
            output_indexes[count] = indexes[i]
            count += 1
            r1 = r2

    return output_arr_indexes[:count]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef max_disjoint32(const long [::1] indexes, const int32_t [::1] starts, const int32_t [::1] ends, int slack):

    cdef int length = len(starts)
    cdef int count = 0
    cdef int r1 = 0
    cdef int l1 = 0
    cdef int r2 = 0

    output_arr_indexes = np.ones(length, dtype=np.int_) * -1
    # output_arr_start = np.ones(length, dtype=np.int32) * -1
    # output_arr_end = np.zeros(length, dtype=np.int32) * -1

    cdef long [::1] output_indexes
    # cdef int32_t [::1] output_start
    # cdef int32_t [::1] output_end

    output_indexes = output_arr_indexes
    # output_start = output_arr_start
    # output_end = output_arr_end

    # output_start[count] = starts[count]
    # output_end[count] = ends[count]
    output_indexes[count] = indexes[count]
    count += 1

    r1 = ends[0]
    for i in range(1, length):
        l1 = starts[i]
        r2 = ends[i]

        if l1 > (r1 + slack):
            # output_start[count] = l1
            # output_end[count] = r2
            output_indexes[count] = indexes[i]
            count += 1
            r1 = r2

    return output_arr_indexes[:count]



cimport cython
from libc.stdint cimport int32_t, int64_t

import numpy as np


def maketiles(indexes, starts, ends, tile_size, preserve_end=False):

    _starts = starts - (starts % tile_size)

    _ends_mod_zero = (ends % tile_size) == 0
    _ends = np.array(len(_starts), dtype=int)
    _ends = np.where(_ends_mod_zero,
                     (ends - 1) - ((ends - 1) % tile_size) + tile_size,
                     ends - (ends % tile_size) + tile_size)

    max_n_tiles = int((np.sum(_ends - _starts + tile_size) / tile_size))
    if max_n_tiles < 0:
        raise Exception("The sum of your chromosome lengths is below 0. Did you switch the start and end columns or try a negative tile size?")

    if starts.dtype == np.int64:
        return maketiles64(indexes, _starts, _ends, max_n_tiles, tile_size)
    elif starts.dtype == np.int32:
        return maketiles32(indexes, _starts, _ends, max_n_tiles, tile_size)
    else:
        raise Exception("Starts/Ends not int64 or int32: " + str(starts.dtype))


@cython.boundscheck(True)
@cython.wraparound(True)
@cython.initializedcheck(True)
cdef maketiles32(const int64_t [::1] indexes, const int32_t [::1] starts, const int32_t [::1] ends, max_n_tiles, int tile_size):

    cdef:
        int nfound = 0
        int i = 0
        int length = len(starts)
        int32_t start, end

    output_arr_indexes = np.ones(max_n_tiles, dtype=np.int64) * -1
    output_arr_start = np.ones(max_n_tiles, dtype=np.int32) * -1
    output_arr_end = np.ones(max_n_tiles, dtype=np.int32) * -1

    cdef long [::1] output_indexes
    cdef int32_t [::1] output_start
    cdef int32_t [::1] output_end

    output_indexes = output_arr_indexes
    output_start = output_arr_start
    output_end = output_arr_end

    for i in range(length):

        start = starts[i]
        end = ends[i]

        if start + tile_size >= end:
            output_indexes[nfound] = indexes[i]
            output_start[nfound] = start
            output_end[nfound] = end
            nfound += 1
        else:
            while start + tile_size <= end:
                output_indexes[nfound] = indexes[i]
                output_start[nfound] = start
                start += tile_size
                output_end[nfound] = start
                nfound += 1

            if start > end:
                output_indexes[nfound] = indexes[i]
                output_start[nfound] = start
                output_end[nfound] = end
                nfound += 1

    return output_arr_indexes[:nfound], output_arr_start[:nfound], output_arr_end[:nfound]



@cython.boundscheck(True)
@cython.wraparound(True)
@cython.initializedcheck(True)
cdef maketiles64(const int64_t [::1] indexes, const int64_t [::1] starts, const int64_t [::1] ends, max_n_tiles, int tile_size):

    cdef:
        int nfound = 0
        int i = 0
        int length = len(starts)
        long start, end

    output_arr_indexes = np.ones(max_n_tiles, dtype=np.int64) * -1
    output_arr_start = np.ones(max_n_tiles, dtype=np.int64) * -1
    output_arr_end = np.ones(max_n_tiles, dtype=np.int64) * -1

    # return np.array([], dtype=np.int64), np.array([], dtype=np.int64), np.array([], dtype=np.int64)
    cdef long [::1] output_indexes
    cdef long [::1] output_start
    cdef long [::1] output_end

    output_indexes = output_arr_indexes
    output_start = output_arr_start
    output_end = output_arr_end

    for i in range(length):

        # print(i)
        start = starts[i]
        end = ends[i]

        if start + tile_size >= end:
            output_indexes[nfound] = indexes[i]
            output_start[nfound] = start
            output_end[nfound] = end
            nfound += 1
        else:
            while start + tile_size <= end:
                output_indexes[nfound] = indexes[i]
                output_start[nfound] = start
                start += tile_size
                output_end[nfound] = start
                nfound += 1

            if start > end:
                output_indexes[nfound] = indexes[i]
                output_start[nfound] = start
                output_end[nfound] = end
                nfound += 1

    return output_arr_indexes[:nfound], output_arr_start[:nfound], output_arr_end[:nfound]




cimport cython
from libc.stdint cimport int32_t

import numpy as np


def makewindows(indexes, starts, ends, window_size):

    _starts = starts - (starts % window_size)
    _ends = ends - (ends % window_size) + window_size

    max_n_windows = int((np.sum(_ends - _starts + window_size) / window_size))

    if starts.dtype == np.int_:
        return makewindows64(indexes, starts, ends, max_n_windows, window_size)
    elif starts.dtype == np.int32:
        return makewindows32(indexes, starts, ends, max_n_windows, window_size)
    else:
        raise Exception("Starts/Ends not int64 or int32: " + str(starts.dtype))


@cython.boundscheck(True)
@cython.wraparound(True)
@cython.initializedcheck(True)
cdef makewindows32(const long [::1] indexes, const int32_t [::1] starts, const int32_t [::1] ends, max_n_windows, int window_size):

    cdef:
        int nfound = 0
        int i = 0
        int length = len(starts)
        int start, end

    output_arr_indexes = np.ones(max_n_windows, dtype=long) * -1
    output_arr_start = np.ones(max_n_windows, dtype=np.int32) * -1
    output_arr_end = np.ones(max_n_windows, dtype=np.int32) * -1

    cdef long [::1] output_indexes
    cdef int32_t [::1] output_start
    cdef int32_t [::1] output_end

    output_indexes = output_arr_indexes
    output_start = output_arr_start
    output_end = output_arr_end

    for i in range(length):

        start = starts[i]
        end = ends[i]

        if start + window_size > end:
            output_indexes[nfound] = indexes[i]
            output_start[nfound] = start
            output_end[nfound] = end
            nfound += 1
            continue


        while start + window_size <= end:
            output_indexes[nfound] = indexes[i]
            output_start[nfound] = start
            start += window_size
            output_end[nfound] = start
            nfound += 1

        if start != end and start + window_size > end:
            output_indexes[nfound] = indexes[i]
            output_start[nfound] = start
            output_end[nfound] = end
            nfound += 1

    return output_arr_indexes[:nfound], output_arr_start[:nfound], output_arr_end[:nfound]



@cython.boundscheck(True)
@cython.wraparound(True)
@cython.initializedcheck(True)
cdef makewindows64(const long [::1] indexes, const long [::1] starts, const long [::1] ends, max_n_windows, int window_size):

    cdef:
        int nfound = 0
        int i = 0
        int length = len(starts)
        long start, end

    output_arr_indexes = np.ones(max_n_windows, dtype=long) * -1
    output_arr_start = np.ones(max_n_windows, dtype=long) * -1
    output_arr_end = np.ones(max_n_windows, dtype=long) * -1

    cdef long [::1] output_indexes
    cdef long [::1] output_start
    cdef long [::1] output_end

    output_indexes = output_arr_indexes
    output_start = output_arr_start
    output_end = output_arr_end

    for i in range(length):

        # print(i)
        start = starts[i]
        end = ends[i]

        if start + window_size > end:
            output_indexes[nfound] = indexes[i]
            output_start[nfound] = start
            output_end[nfound] = end
            nfound += 1
            continue


        while start + window_size <= end:
            output_indexes[nfound] = indexes[i]
            output_start[nfound] = start
            start += window_size
            output_end[nfound] = start
            nfound += 1

        if start != end and start + window_size > end:
            output_indexes[nfound] = indexes[i]
            output_start[nfound] = start
            output_end[nfound] = end
            nfound += 1

    return output_arr_indexes[:nfound], output_arr_start[:nfound], output_arr_end[:nfound]




cimport cython
from libc.stdint cimport int32_t

import numpy as np


def find_introns(gene_starts, gene_ends, gene_ids,
                 exon_starts, exon_ends, exon_gene_ids):

    if gene_starts.dtype == np.int_:
        return find_introns64(gene_starts, gene_ends, gene_ids, exon_starts, exon_ends, exon_gene_ids)
    elif gene_starts.dtype == np.int32:
        return find_introns32(gene_starts, gene_ends, gene_ids, exon_starts, exon_ends, exon_gene_ids)
    else:
        raise Exception("Starts/Ends not int64 or int32: " + str(gene_starts.dtype))


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef find_introns64(const long [::1] gene_starts, const long [::1] gene_ends, const long [::1] gene_ids,
                     const long [::1] exon_starts, const long [::1] exon_ends, const long [::1] x_gene_ids):

    cdef:
        exon_length = len(exon_starts)
        length = 2 * exon_length
        int n = 0
        int i = 0
        int j = 0
        long gene_start, gene_end, gene_id
        long [::1] output_start
        long [::1] output_end
        long [::1] output_id

    output_arr_start = np.ones(length, dtype=np.int_) * -1
    output_arr_end = np.zeros(length, dtype=np.int_) * -1
    output_arr_id = np.zeros(length, dtype=np.int_) * -1

    output_id = output_arr_id
    output_start = output_arr_start
    output_end = output_arr_end

    for i in range(len(gene_starts)):

        gene_id = gene_ids[i]
        gene_start = gene_starts[i]
        gene_end = gene_ends[i]

        if gene_start != exon_starts[j]:
            output_start[n] = gene_start
            output_end[n] = exon_starts[j]
            output_id[n] = gene_id
            n += 1

        j += 1

        while j < exon_length and gene_id == x_gene_ids[j]:

            output_start[n] = exon_ends[j - 1]
            output_end[n] = exon_starts[j]
            output_id[n] = gene_id
            n += 1
            j += 1

        if exon_ends[j - 1] != gene_end:

            output_start[n] = exon_ends[j - 1]
            output_end[n] = gene_end
            output_id[n] = gene_id
            n += 1

    return output_arr_start[:n], output_arr_end[:n], output_arr_id[:n]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef find_introns32(const int32_t [::1] gene_starts, const int32_t [::1] gene_ends, const long [::1] gene_ids,
                     const int32_t [::1] exon_starts, const int32_t [::1] exon_ends, const long [::1] x_gene_ids):

    cdef:
        exon_length = len(exon_starts)
        length = 2 * exon_length
        int n = 0
        int i = 0
        int j = 0
        int gene_start, gene_end, gene_id
        int32_t [::1] output_start
        int32_t [::1] output_end
        int32_t [::1] output_id

    output_arr_start = np.ones(length, dtype=np.int32) * -1
    output_arr_end = np.zeros(length, dtype=np.int32) * -1
    output_arr_id = np.zeros(length, dtype=np.int32) * -1

    output_id = output_arr_id
    output_start = output_arr_start
    output_end = output_arr_end

    for i in range(len(gene_starts)):

        gene_id = gene_ids[i]
        gene_start = gene_starts[i]
        gene_end = gene_ends[i]
        # print("-----" * 5)
        # print("gene_id", gene_id)
        # print("gene_start", gene_start)
        # print("gene_end", gene_end)
        # print("j", j)
        # print()
        # assert gene_ids[i] == x_gene_ids[j], "gene_exon_not_equal"

        # print("gene_start != exon_starts[j]", gene_start != exon_starts[j])

        # assert gene_start <= exon_starts[j], "gene_start not smaller or equal to exon_end: {}, {}".format(gene_start, exon_starts[j])

        if gene_start != exon_starts[j]:
            # print("exon_start != gene_start")
            # print("insert start", gene_start)
            # print("insert end", exon_starts[j])
            # print()
            output_start[n] = gene_start
            output_end[n] = exon_starts[j]
            output_id[n] = gene_id
            n += 1

        j += 1

        # print("gene_start != exon_starts[j]", gene_start != exon_starts[j])
        # print("j", j)
        # if j < exon_length:
        #     print("gene_id == x_gene_ids[j]", gene_id == x_gene_ids[j], gene_id, x_gene_ids[j])
        while j < exon_length and gene_id == x_gene_ids[j]:
            # if exon_ends[j - 1] == 1217804:
            #     print(gene_id)
            #     raise
            # print("insert start", exon_ends[j - 1])
            # print("insert end", exon_starts[j])
            # print()

            output_start[n] = exon_ends[j - 1]
            output_end[n] = exon_starts[j]
            output_id[n] = gene_id
            n += 1
            j += 1

        # if j < exon_length:
            # print("while ended here:", gene_id, x_gene_ids[j])

        if exon_ends[j - 1] != gene_end:
            # print("exon_ends[j - 1] != gene_end", exon_ends[j - 1] != gene_end)
            # print("insert start", exon_ends[j - 1])
            # print("insert end", gene_end)
            # print()

            output_start[n] = exon_ends[j - 1]
            output_end[n] = gene_end
            output_id[n] = gene_id
            n += 1

    return output_arr_start[:n], output_arr_end[:n], output_arr_id[:n]





cimport cython
from libc.stdint cimport int32_t

import numpy as np


def cluster_by(starts, ends, ids, slack=0):

    if starts.dtype == np.int_:
        return cluster_by64(starts, ends, ids.astype(np.int_), slack)
    elif starts.dtype == np.int32:
        return cluster_by32(starts, ends, ids.astype(np.int32), slack)
    else:
        raise Exception("Starts/Ends not int64 or int32: " + str(starts.dtype))


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef cluster_by64(const long [::1] starts, const long [::1] ends, const long [::1] ids, int slack):

    cdef long min_start = starts[0]
    cdef long max_end = ends[0]
    cdef int i = 0
    cdef int current_id = ids[0]
    cdef int last_id = ids[0]
    cdef int n_clusters = 1
    cdef int length = len(starts)

    output_arr_ids = np.ones(length, dtype=np.int_) * -1

    cdef long [::1] output_ids

    output_ids = output_arr_ids

    for i in range(length):
        current_id = ids[i]
        if (not (starts[i] - slack) <= max_end) or current_id != last_id:
            n_clusters += 1
            output_ids[i] = n_clusters
            min_start = starts[i]
            max_end = ends[i]
        else:
            output_ids[i] = n_clusters
            if ends[i] > max_end:
                max_end = ends[i]
        last_id = current_id

    # if n_clusters != length + 1:
    #     output_arr_ids[i] = n_clusters
    #     n_clusters += 1

    return output_arr_ids


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef cluster_by32(const int32_t [::1] starts, const int32_t [::1] ends, const int32_t [::1] ids, int slack):

    cdef int min_start = starts[0]
    cdef int max_end = ends[0]
    cdef int i = 0
    cdef int current_id = ids[0]
    cdef int last_id = ids[0]
    cdef int n_clusters = 1
    cdef int length = len(starts)

    output_arr_ids = np.ones(length, dtype=np.int32) * -1

    cdef int32_t [::1] output_ids

    output_ids = output_arr_ids

    for i in range(length):
        # print("i " * 5, i)
        current_id = ids[i]
        if (not (starts[i] - slack) <= max_end) or current_id != last_id:
            n_clusters += 1
            output_ids[i] = n_clusters
            min_start = starts[i]
            max_end = ends[i]
        else:
            # print("else")
            output_ids[i] = n_clusters
            if ends[i] > max_end:
                max_end = ends[i]
        last_id = current_id

    # if n_clusters != length + 1:
    #     output_arr_ids[i] = n_clusters
    #     n_clusters += 1

    return output_arr_ids



import pkg_resources

from sorted_nearest.src.annotate_clusters import annotate_clusters  # NOQA: F401
from sorted_nearest.src.cluster_by import cluster_by  # NOQA: F401
from sorted_nearest.src.clusters import find_clusters  # NOQA: F401
from sorted_nearest.src.introns import find_introns  # NOQA: F401
from sorted_nearest.src.k_nearest import k_nearest_next_nonoverlapping, k_nearest_previous_nonoverlapping  # NOQA: F401
from sorted_nearest.src.k_nearest_ties import get_all_ties, get_different_ties  # NOQA: F401
from sorted_nearest.src.max_disjoint_intervals import max_disjoint  # NOQA: F401
from sorted_nearest.src.merge_by import merge_by  # NOQA: F401
from sorted_nearest.src.sorted_nearest import (  # NOQA: F401
    nearest_next_nonoverlapping,
    nearest_nonoverlapping,
    nearest_previous_nonoverlapping,
)
from sorted_nearest.src.tiles import maketiles  # NOQA: F401
from sorted_nearest.src.windows import makewindows  # NOQA: F401

__version__ = pkg_resources.get_distribution("sorted_nearest").version



cimport cython
from libc.stdint cimport int32_t, int64_t

import numpy as np


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cpdef get_all_ties(const int64_t[::1] lx, const int64_t[::1] ids, const int64_t[::1] dist, int k):

    """all fetches until you have k nearest intervals and then all intervals with same distance."""

    # for each index
    # want to keep until we have k
    # then keep all with same distance
    cdef:
        int i = 0
        int first_i = 0
        int current_found = 0
        int nfound = 0
        int current_id = ids[0]
        int last_id = -1
        int length = len(lx)
        int current_dist = -1
        int last_dist = -1

    arr_lidx = np.ones(length, dtype=np.int_) * -1

    cdef long [::1] lidx

    lidx = arr_lidx

    while i < length:

        # print("------" * 10)
        # print("i", i)
        # print("ids[i]", ids[i])
        # print("last_id", last_id)
        # print("current_found", current_found)
        # print("nfound", nfound)
        # print("length", length)
        first_i = i

        if ids[i] != last_id:
            # we have come to a new id, need to start counting again
            # print("** if **")
            # print("i", i, "length", length, "ids[i]", ids[i], "last_id", last_id, "dist[i]", last_dist, "current_found", current_found)

            current_found = 0
            current_id = ids[i]
            last_id = ids[i]
            # continue

        # while have not found k different ties, continue
        while i < length and ids[i] == last_id and current_found < k:
            # print("**first while**")
            # print("i", i, "length", length, "ids[i]", ids[i], "last_id", last_id, "dist[i]", last_dist, "current_found", current_found)
            lidx[nfound] = lx[i]
            # print("lidx", arr_lidx)
            last_dist = dist[i]
            nfound += 1
            current_found += 1
            i += 1

        # while we have not found all ties with same length as last included
        while i < length and ids[i] == last_id and dist[i] == last_dist:
            # print("**second while**")
            # print("i", i, "length", length, "ids[i]", ids[i], "last_id", last_id, "dist[i]", last_dist, "current_found", current_found)
            lidx[nfound] = lx[i]
            # print("lidx", arr_lidx)
            nfound += 1
            i += 1

        if i == first_i:
            i += 1

        # print(arr_lidx)

    return arr_lidx[:nfound]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cpdef get_different_ties(const int64_t[::1] lx, const int64_t[::1] ids, const int64_t[::1] dist, int k):

    """fetch all ties until you have all intervals with k distances"""

    # print("lx", list(lx))
    # print("ids", list(ids))
    # print("dist", list(dist))

    # for each index
    # want to keep until we have k
    # then keep all with same distance
    cdef:
        int i = 0
        int first_i = 0
        int current_found = 0
        int nfound = 0
        int current_id = ids[0]
        int last_id = -1
        int length = len(lx)
        int current_dist = -1
        int last_dist = -1

    arr_lidx = np.ones(length, dtype=np.int_) * -1

    cdef long [::1] lidx

    lidx = arr_lidx

    # Chromosome  Start  End Name  Score Strand  __k__  __IX__  Start_b  End_b Name_b  Score_b Strand_b  Distance
    # 0       chr1      1    2    a      0      -      2       0        1      2      a        0        -         0
    # 1       chr1      1    2    a      0      -      2       0        1      2      a        0        -         0
    # 2       chr1      1    2    a      0      -      2       0        1      2      a        0        -         0

    while i < length:

        # print("i", i)
        # print("ids[i]", ids[i])
        # print("last_id", last_id)
        # print("current_found", current_found)
        # print("nfound", nfound)
        # print("length", length)
        first_i = i

        if ids[i] != last_id:
            # we have come to a new id, need to start counting again


            current_found = 0
            current_id = ids[i]
            last_id = ids[i]
            continue

        # while have not found k different ties, continue
        while i < length and ids[i] == last_id and current_found < k:
            # print("**first while**")
            # print("i", i, "length", length, "ids[i]", ids[i], "last_id", last_id, "dist[i]", last_dist, "current_found", current_found)
            lidx[nfound] = lx[i]
            # print("lidx", arr_lidx)
            last_dist = dist[i]
            nfound += 1
            i += 1

            if i < length and dist[i] != last_dist:
                current_found += 1

        # while we have not found all ties with same length as last included
        while i < length and ids[i] == last_id and dist[i] == last_dist:
            # print("**second while**")
            # print("i", i, "length", length, "ids[i]", ids[i], "last_id", last_id, "dist[i]", last_dist, "current_found", current_found)
            lidx[nfound] = lx[i]
            # print("lidx", arr_lidx)
            nfound += 1
            i += 1

        if i == first_i:
            i += 1

        # print(arr_lidx)

    return arr_lidx[:nfound]



from distutils.core import setup
from setuptools import Extension
from Cython.Build import cythonize


extensions = [
    Extension(
        "sorted_nearest.src.sorted_nearest",
        ["sorted_nearest/src/sorted_nearest.pyx"],
    ),
    Extension(
        "sorted_nearest.src.max_disjoint_intervals",
        ["sorted_nearest/src/max_disjoint_intervals.pyx"],
    ),
    Extension(
        "sorted_nearest.src.k_nearest",
        ["sorted_nearest/src/k_nearest.pyx"],
    ),
    Extension(
        "sorted_nearest.src.k_nearest_ties",
        ["sorted_nearest/src/k_nearest_ties.pyx"],
    ),
    Extension(
        "sorted_nearest.src.clusters",
        ["sorted_nearest/src/clusters.pyx"],
    ),
    Extension(
        "sorted_nearest.src.annotate_clusters",
        ["sorted_nearest/src/annotate_clusters.pyx"],
    ),
    Extension(
        "sorted_nearest.src.cluster_by",
        ["sorted_nearest/src/cluster_by.pyx"],
    ),
    Extension(
        "sorted_nearest.src.merge_by",
        ["sorted_nearest/src/merge_by.pyx"],
    ),
    Extension(
        "sorted_nearest.src.introns",
        ["sorted_nearest/src/introns.pyx"],
    ),
    Extension(
        "sorted_nearest.src.windows",
        ["sorted_nearest/src/windows.pyx"],
    ),
    Extension(
        "sorted_nearest.src.tiles",
        ["sorted_nearest/src/tiles.pyx"],
    ),
]

setup(ext_modules=cythonize(extensions, language_level=3))



"find clusters is what is called merge in pyranges"

cimport cython
from libc.stdint cimport int32_t

import numpy as np


def find_clusters(starts, ends, slack):

    if starts.dtype == np.int_:
        return find_clusters64(starts, ends, slack)
    elif starts.dtype == np.int32:
        return find_clusters32(starts, ends, slack)
    else:
        raise Exception("Starts/Ends not int64 or int32: " + str(starts.dtype))


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef find_clusters64(const long [::1] starts, const long [::1] ends, int slack):

    cdef long min_start = starts[0]
    cdef long max_end = ends[0]
    cdef int i = 0
    cdef int intervals_in_cluster = 0
    cdef int n_clusters = 0
    cdef int length = len(starts)

    output_arr_start = np.ones(length, dtype=np.int_) * -1
    output_arr_end = np.zeros(length, dtype=np.int_) * -1
    output_arr_number = np.zeros(length, dtype=np.int_) * -1

    cdef long [::1] output_start
    cdef long [::1] output_end
    cdef long [::1] output_number

    output_start = output_arr_start
    output_end = output_arr_end
    output_number = output_arr_number

    for i in range(length):
        if not (starts[i] - slack) <= max_end:
            output_start[n_clusters] = min_start
            output_end[n_clusters] = max_end
            output_number[n_clusters] = intervals_in_cluster
            min_start = starts[i]
            max_end = ends[i]
            intervals_in_cluster = 1
            n_clusters += 1
        else:
            intervals_in_cluster += 1
            if ends[i] > max_end:
                max_end = ends[i]

    if n_clusters < length and output_arr_start[n_clusters] != min_start:
        output_arr_start[n_clusters] = min_start
        output_arr_end[n_clusters] = max_end
        output_number[n_clusters] = intervals_in_cluster
        n_clusters += 1

    return output_arr_start[:n_clusters], output_arr_end[:n_clusters], output_arr_number[:n_clusters]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef find_clusters32(const int32_t [::1] starts, const int32_t [::1] ends, int slack):

    cdef int min_start = starts[0]
    cdef int max_end = ends[0]
    cdef int i = 0
    cdef int n_clusters = 0
    cdef int intervals_in_cluster = 0
    cdef int length = len(starts)

    output_arr_start = np.ones(length, dtype=np.int32) * -1
    output_arr_end = np.zeros(length, dtype=np.int32) * -1
    output_arr_number = np.zeros(length, dtype=np.int32) * -1

    cdef int32_t [::1] output_start
    cdef int32_t [::1] output_end
    cdef int32_t [::1] output_number

    output_start = output_arr_start
    output_end = output_arr_end
    output_number = output_arr_number

    for i in range(length):
        if not (starts[i] - slack) <= max_end:
            output_start[n_clusters] = min_start
            output_end[n_clusters] = max_end
            output_number[n_clusters] = intervals_in_cluster
            min_start = starts[i]
            max_end = ends[i]
            intervals_in_cluster = 1
            n_clusters += 1
        else:
            intervals_in_cluster += 1
            if ends[i] > max_end:
                max_end = ends[i]

    if n_clusters < length and output_arr_start[n_clusters] != min_start:
        output_arr_start[n_clusters] = min_start
        output_arr_end[n_clusters] = max_end
        output_number[n_clusters] = intervals_in_cluster
        n_clusters += 1

    return output_arr_start[:n_clusters], output_arr_end[:n_clusters], output_arr_number[:n_clusters]



# cdef extern from "sorted_nearest/src/utarray.h":

#     ctypedef struct UT_icd:
#         pass

#     ctypedef struct UT_array:
#         pass

#     const UT_icd ut_int_icd

#     void utarray_new(UT_array *a, UT_icd *icd)
#     int utarray_len(UT_array *a)
#     int* utarray_eltptr(UT_array *a, int j)
#     void utarray_push_back(UT_array *a, void *p)
#     void utarray_free(UT_array *a)

# cdef extern from "limits.h":
#     int INT_MAX
